/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "buffertable.h"

#include <algorithm>
#include <sstream>

namespace bitGtk
  {

  BufferTable::BufferTable():
      Gtk::HBox(),
      m_buffer(NULL),
      m_display_bit_width(0),
      m_highlighter_enabled(true),
      m_bg_color("black"),
      m_table(NULL)
  {

    m_tablebox.modify_bg(Gtk::STATE_NORMAL, m_bg_color);
    m_tablebox.set_border_width(3);
    create_table();
    m_tableboxouter.modify_bg(Gtk::STATE_NORMAL, m_bg_color);
    m_tableboxouter.add(m_tablebox);
    pack_start(m_tableboxouter);

    add_color_set("aliceblue");
    add_color_set("lightgreen");
    add_color_set("moccasin");
    add_color_set("thistle");
    add_color_set("beige");
    add_color_set("lightpink");
    add_color_set("lightyellow");
    add_color_set("orange");
    add_color_set("powderblue");
    add_color_set("plum");
    add_color_set("navajowhite");
  }

  BufferTable::~BufferTable()
  {}

  void BufferTable::set_buffer( bit::RecordBuffer& buffer )
  {
    m_buffer = &buffer;
    m_record = buffer.record();
    setup_table();
    load_all_field_values();
    buffer.signal_field_changed().connect(sigc::mem_fun(*this, &BufferTable::on_field_value_changed));
    buffer.signal_record_changed().connect(sigc::mem_fun(*this, &BufferTable::on_record_changed));
  }

  void BufferTable::on_record_changed() {
    load_all_field_values();
  }

  void BufferTable::set_bit_width( size_t w )
  {
    m_display_bit_width = w;
    setup_table();
  }

  void BufferTable::set_record( bit::FieldBase::pointer record )
  {
    m_buffer = NULL;
    m_record = record;
    setup_table();
  }

  void BufferTable::create_table( )
  {
    if (m_table != NULL)
      m_tablebox.remove();

    m_table = Gtk::manage(new Gtk::Table());
    m_table->set_homogeneous();
    m_table->set_spacings(3);
    m_tablebox.add(*m_table);
  }

  void BufferTable::setup_table( )
  {
    std::ostringstream text;
    size_t maximum;
    size_t rows, current_position, current_row;
    std::list<std::pair<Gdk::Color, Gdk::Color> >::iterator color = m_colors.begin();
    bit::Data data;
    std::string tooltip;

    // check to make sure we have set the record pointer
    if (!m_record)
      return;

    // start with a fresh, clean table and destroy the old one if necessary
    create_table();
    m_tablebox.modify_bg(Gtk::STATE_NORMAL, m_bg_color);
    m_tableboxouter.modify_bg(Gtk::STATE_NORMAL, m_bg_color);

    // clear our pointer storage data structures
    m_fieldbox.clear();

    // find the maximum bit field size to display
    maximum = m_record->start(bit::BITS) + m_record->length(bit::BITS);

    // calculate how many rows we will be displaying and adjust the table accordingly
    if (m_display_bit_width > 0)
      rows = maximum / m_display_bit_width;
    else
      rows = 1;

    // iterate through all the fields creating as many boxes as are necessary
    current_position = 0;
    current_row = 0;

    create_fieldboxes(m_record);

    m_table->resize_children();
  }

  void BufferTable::create_fieldboxes(bit::FieldBase::pointer fb)
  {
    size_t remaining_bits, current_row=0;
    size_t left, right;
    static std::list<std::pair<Gdk::Color, Gdk::Color> >::iterator color = m_colors.begin();

    if (!fb)
      return;

    if (fb->fields())
      {
        for (bit::FieldBase::iterator iter = fb->begin(); iter != fb->end(); iter++)
          create_fieldboxes(iter.pointer());
      }
    else
      {

        // start off with all the bits needing a display box
        remaining_bits = fb->length(bit::BITS);

        // find the proper row for this label to start on
        if (m_display_bit_width > 0)
          current_row = fb->start(bit::BITS) / m_display_bit_width;

        // find the left and right attachment points of the hbox and attach it to table
        if (m_display_bit_width > 0)
          {
            left = fb->start(bit::BITS) % m_display_bit_width;
            right = std::min(left + remaining_bits, m_display_bit_width);
          }
        else
          {
            left = fb->start(bit::BITS);
            right = left + remaining_bits;
          }

        add_fieldbox(fb, left, right, current_row, current_row+1, color->first, color->second);

        // calculate how many bits we have left in this field
        remaining_bits -= right - left;

        // if our display width is too small for the whole field we need more display
        // boxes for the remainder of the field
        // fortunately, they all start on the left boundary
        left = 0;

        // create the remaining row labels (if any bits remain)
        while (remaining_bits)
          {
            // move to the next row
            current_row++;

            // we know the left attachment point is 0, but still need to calculate the right
            right = std::min(left + remaining_bits, m_display_bit_width);

            add_fieldbox(fb, left, right, current_row, current_row+1, color->first, color->second);

            // calculate how many bits we might still have left to add a display box for
            remaining_bits -= right - left;

          }

        // if we have a color palette move to the next color or start over
        if (m_colors.size() > 1)
          {
            color++;
            if (color == m_colors.end())
              color = m_colors.begin();
          }
      }
  }

  void BufferTable::add_fieldbox(bit::FieldBase::pointer field, unsigned left, unsigned right, unsigned top, unsigned bottom, Gdk::Color bg, Gdk::Color hl)
  {
    std::ostringstream text;
    std::string tooltip;

    // create the hbox to contain the element and buffer labels and the backing
    // highlight box
    m_fieldbox[field].m_highlight_boxes.push_back( Gtk::manage(new Gtk::VBox()) );

    // create the element and buffer labels and pack them into event boxes
    // the event boxes are necessary to show the background color and to capture
    // events for future enhancements
    // also add them to the vectors
    m_fieldbox[field].m_field_labels.push_back( Gtk::manage(new Gtk::Label()) );
    m_fieldbox[field].m_field_event_boxes.push_back( Gtk::manage(new Gtk::EventBox()) );
    m_fieldbox[field].m_field_event_boxes.back()->add
    (*(m_fieldbox[field].m_field_labels.back()));

    m_fieldbox[field].m_value_labels.push_back( Gtk::manage(new Gtk::Label()) );
    m_fieldbox[field].m_value_event_boxes.push_back( Gtk::manage(new Gtk::EventBox()) );
    m_fieldbox[field].m_value_event_boxes.back()->add
    (*(m_fieldbox[field].m_value_labels.back()));

    // add the element and buffer event boxes to the hbox
    m_fieldbox[field].m_highlight_boxes.back()->pack_start(*(m_fieldbox[field].m_field_event_boxes.back()));
    m_fieldbox[field].m_highlight_boxes.back()->pack_start(*(m_fieldbox[field].m_value_event_boxes.back()));

    // add tooltips if there is a description
    if (field->description().size() > 0)
      {
        tooltip = field->name(field->depth()-1) + "\n\n" + field->description();
        m_tooltips.set_tip(*(m_fieldbox[field].m_field_event_boxes.back()), tooltip);
        m_tooltips.set_tip(*(m_fieldbox[field].m_value_event_boxes.back()), tooltip);
      }

    m_fieldbox[field].m_field_event_boxes.back()->signal_enter_notify_event().connect(sigc::bind(sigc::mem_fun(*this, &BufferTable::on_enter), field));
    m_fieldbox[field].m_value_event_boxes.back()->signal_enter_notify_event().connect(sigc::bind(sigc::mem_fun(*this, &BufferTable::on_enter), field));
    m_fieldbox[field].m_field_event_boxes.back()->signal_leave_notify_event().connect(sigc::bind(sigc::mem_fun(*this, &BufferTable::on_leave), field));
    m_fieldbox[field].m_value_event_boxes.back()->signal_leave_notify_event().connect(sigc::bind(sigc::mem_fun(*this, &BufferTable::on_leave), field));

    m_fieldbox[field].m_field_event_boxes.back()->modify_bg(Gtk::STATE_NORMAL, bg);
    m_fieldbox[field].m_field_event_boxes.back()->modify_bg(Gtk::STATE_ACTIVE, hl);
    m_fieldbox[field].m_value_event_boxes.back()->modify_bg(Gtk::STATE_NORMAL, bg);
    m_fieldbox[field].m_value_event_boxes.back()->modify_bg(Gtk::STATE_ACTIVE, hl);

    m_table->attach(*(m_fieldbox[field].m_highlight_boxes.back()), left, right, top, bottom);

    // fill the element label text
    text << " " << field->name( field->depth()-1 ) << " \n " << left << "-" << right-1 << " ";
    m_fieldbox[field].m_field_labels.back()->set_text(text.str());
    m_fieldbox[field].m_field_labels.back()->set_justify(Gtk::JUSTIFY_CENTER);
  }

  void BufferTable::on_field_value_changed( bit::FieldBuffer fieldbuffer )
  {
    std::vector<Gtk::Label*>::iterator i;

    std::string text;
    text = std::string("<i>") + fieldbuffer.data().hex_string() + std::string("</i>");
    for ( i = m_fieldbox[fieldbuffer.field()].m_value_labels.begin(); i != m_fieldbox[fieldbuffer.field()].m_value_labels.end(); i++)
      {
        (*i)->set_use_markup();
        (*i)->set_markup(text);
      }
  }

  void BufferTable::set_field_value( bit::FieldBase::pointer field )
  {
    on_field_value_changed( bit::FieldBuffer( *m_buffer, field ) );
  }

  void BufferTable::load_all_field_values()
  {
    if (! m_buffer )
      return;

    load_all_field_values( m_buffer->record() );
  }

  void BufferTable::load_all_field_values(bit::FieldBase::pointer field)
  {
    if ( field->fields() > 0)
      for (bit::FieldBase::iterator iter = field->begin(); iter != field->end(); iter++)
        load_all_field_values( iter.pointer() );
    else
      set_field_value( field );
  }

  void BufferTable::enable_tooltips( bool b )
  {
    if (b)
      m_tooltips.enable();
    else
      m_tooltips.disable();
    setup_table();
  }

  void BufferTable::enable_highlighter( bool b )
  {
    m_highlighter_enabled = b;
    setup_table();
  }

  void BufferTable::set_bg_color( std::string s )
  {
    m_bg_color.set(s);
    setup_table();
  }

  bool BufferTable::on_enter( GdkEventCrossing *, bit::FieldBase::pointer field )
  {
    if (! m_highlighter_enabled)
      return false;

    std::vector<Gtk::VBox*>::iterator i;
    for (i = m_fieldbox[field].m_highlight_boxes.begin(); i != m_fieldbox[field].m_highlight_boxes.end(); i++)
      (*i)->set_state(Gtk::STATE_ACTIVE);

    return true;
  }

  bool BufferTable::on_leave( GdkEventCrossing *, bit::FieldBase::pointer field )
  {
    if (! m_highlighter_enabled)
      return false;

    std::vector<Gtk::VBox*>::iterator i;
    for (i = m_fieldbox[field].m_highlight_boxes.begin(); i != m_fieldbox[field].m_highlight_boxes.end(); i++)
      (*i)->set_state(Gtk::STATE_NORMAL);

    return true;
  }

  void BufferTable::add_color_set( std::string normal_color, std::string highlight_color )
  {
    std::pair<Gdk::Color, Gdk::Color> colorset;
    colorset.first.set(normal_color);
    colorset.second.set(highlight_color);
    m_colors.push_back(colorset);
  }

  void BufferTable::clear_color_sets( )
  {
    m_colors.clear();
  }

}
